{-# OPTIONS_GHC -Wall         #-}
module HarmTrace.Audio.BeatChroma where

import HarmTrace.Audio.Parser
import HarmTrace.Audio.ChordTypes
import HarmTrace.Audio.Utils
import HarmTrace.Base.MusicRep

-- import Text.Printf (printf)
import Data.List (sortBy, groupBy)
import Data.Ord (comparing)

import Numeric.LinearAlgebra -- http://hackage.haskell.org/package/hmatrix 
import qualified Numeric.GSL.Statistics as GSL hiding (mean)

--------------------------------------------------------------------------------
-- Local Parameters
--------------------------------------------------------------------------------

-- The cutOffProbability is the value that determines the length of the 
-- probChord list (a sorted list with normalised distances to the chroma vector)
cutOffProbability      :: NumData
cutOffProbability      = 0.92

-- The maximum number of probabilities
maxProbChordListLength :: Int
maxProbChordListLength = 6

--------------------------------------------------------------------------------
-- Matching chords and chroma
--------------------------------------------------------------------------------  

segmentByKey :: [TimedData Key] -> [TimedData [ProbChord]] -> [ProbChordSeg]
segmentByKey []       _    = error "segmentByKey: empty key list"
segmentByKey [k]      chds = [Segment (getData k) chds]
segmentByKey (k : ks) chds = let (seg,cs) = span ((<= (offset k)) . offset) chds
  in Segment (getData k) seg : segmentByKey ks cs
  
groupPChord :: [[ProbChord]] -> [[[ProbChord]]]
groupPChord = groupBy probChordEqPerm

-- a function for comparing list of probchords, two list are considered equal
-- if the chords and their order match, the probabilities are ignored
probChordEq :: [ProbChord] -> [ProbChord] -> Bool
probChordEq []     []     = True
probChordEq (p:pc) (q:qs) = chordLab p == chordLab q && probChordEq pc qs
probChordEq _      _      = False

-- a function for comparing list of probchords, two list are considered equal
-- if the chords and their order match, the probabilities are ignored
probChordEqPerm :: [ProbChord] -> [ProbChord] -> Bool
probChordEqPerm p q = eqPerm (map chordLab p) (map chordLab q)
eqPerm :: [ChordLabel] -> [ChordLabel] -> Bool
eqPerm []     _  = True
eqPerm (p:pc) qs = p `elem` qs && eqPerm pc qs
-- eqPerm (p:pc) qs = ( qs == [noneLabel] || p `elem` qs ) && eqPerm pc qs

-- syncronises the last beat and key frame to match the last chord frame
syncEndings ::  [ChordinoLine] -> [NumData] -> [ChordinoLine]
            -> ([ChordinoLine],   [NumData],   [ChordinoLine])
syncEndings chrm beat key = 
  let endTime   = time $ last chrm
      key'      = takeWhile ((< endTime).time) key 
      (l,[lst]) = splitAt ((length key') - 1) key'
  in (chrm, takeWhile (< endTime) beat ++ [endTime]
          , l                          ++ [lst{time = endTime}])

-- creates "timed" blocks, given beat and beat aligned data 
-- (used by simpleannotator)
addTimeInfo :: BeatTrackerData -> [a] -> [TimedData a]
addTimeInfo offs blcks = zipWith3 TimedData blcks ons offs where
  ons = 0 : offs 

--------------------------------------------------------------------------------
-- Matrix Functions for matching chords
-------------------------------------------------------------------------------- 

createChordRanks :: BeatChroma -> [[ProbChord]]
createChordRanks = 
  map (selectTop . normalize .sortTake . matchCDictionary) 
      . meanBeatSyncVectors where
    sortTake, normalize :: [ProbChord] -> [ProbChord]
    sortTake          = take maxProbChordListLength . sortBy (comparing prob)
    normalize l@(h:_) = map (\(ProbChord c p) -> (ProbChord c (prob h / p))) l 
    normalize []      = []
    selectTop l -- selects the everything with a propability > x
      | null s       = none 
      -- so far, this had not possitive effect
      -- | length s > maxProbChordListLength = none 
      | otherwise    = s  
          where s    = takeWhile ((> cutOffProbability) . prob) l 
                none = [ProbChord noneLabel 0.0]

-- takes the mean of every "beat block" and these Vectors as one Matrix
-- Each row of this matrix corresponds to the chroma within one beat
meanBeatSyncVectors :: BeatChroma -> [Vector NumData] -- [[ChordinoLine]]
meanBeatSyncVectors = map mean . beatSyncMatrix

-- takes the median of every "beat block" and these Vectors as one Matrix
-- Each row of this matrix corresponds to the chroma within one beat
-- N.B. does not perform as well as meanBeatSyncVectors
-- medianBeatSyncVectors :: BeatChroma -> [Vector NumData]
-- medianBeatSyncVectors = 
   -- map (fromList . map GSL.median . toColumns) . beatSyncMatrix
 
-- creates a list of matrices, in which each matrix corresponds to the
-- collection of chroma vectors within one beat (drop the time stamp)
beatSyncMatrix :: BeatChroma -> [Matrix NumData]
beatSyncMatrix dat = map (dropColumns 1 . toChromaMatrix) dat

-- converst a ChordinoData into a Matrix
toChromaMatrix :: ChordinoData -> Matrix NumData
toChromaMatrix = fromLists . map mergeLine where
  mergeLine :: ChordinoLine -> [NumData]
  mergeLine (ChordinoLine tm bs tb) = tm : bs ++ tb    
  
-- matches all 12 tranposisition of a chord structure with a chorma vector  
matchCDictionary :: Vector NumData -> [ProbChord]
matchCDictionary v = map (matchStruct v) chordDictionary

-- calculate an Euclidean (PNorm2) norm
matchStruct :: Vector NumData -> ChordCand -> ProbChord
matchStruct chroma cc@(ChordCand r sh _cs) = 
  -- Chord root shorthand degrees description_str repetitions
  ProbChord (Chord r sh [] 0 1) (pnorm PNorm2 (chroma - bassTrebStruct cc))
            -- (correlation chroma (bassTrebStruct cc))

-- creates a "matchable" 24D vector given a ChordStruct
bassTrebStruct :: ChordCand -> Vector NumData
bassTrebStruct (ChordCand _r _sh cs) = fromList (cs ++ cs)
-- does not seem to improve performance
{-  | sh == None  = fromList (cs ++ cs)
  | otherwise   = fromList (bass ++ cs) where
    bass          = (f before) ++ 1 : (f $ tail after)
    (before,after)= splitAt (diaNatToSemi r) cs
    f             = map (* 0.5)
-}                    
--------------------------------------------------------------------------------
-- The Chord Dictionary
--------------------------------------------------------------------------------

-- the chord dictionary of all chords that are matched
chordDictionary :: [ChordCand]
chordDictionary = concatMap transpose12 [minBound..] -- all shorthands :-)
  
-- takes a ChordStruct and returns a list containing the 12
-- tranposed verions (including the original) of the ChordStruct  
transpose12 :: Shorthand -> [ChordCand]
transpose12 shand 
  | null cstruct  = []
  | shand == None = [ChordCand (Note Nothing N) None cstruct]
  | otherwise     = transpose12' 11 shand cstruct where
      cstruct              = (shortHandToChordStruct shand)
      transpose12' :: Int -> Shorthand -> ChordStruct -> [ChordCand]
      transpose12' 0 sh cs = [ChordCand (head chromaPC) sh cs]
      transpose12' n sh cs = ChordCand (chromaPC!!(12-n)) sh (shift n cs) 
                             : transpose12' (n-1) sh cs

shortHandToChordStruct :: Shorthand -> ChordStruct
shortHandToChordStruct sh = case sh of 
  --        0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10,  11
  --        C,  Db, D,  Eb, E,  F,  F#, G,  Ab, A,  Bb,  B
  Maj   -> [1,  0,  0,  0,  1,  0,  0,  1,  0,  0,  0,   0 ] 
  Min   -> [1,  0,  0,  1,  0,  0,  0,  1,  0,  0,  0,   0 ]
  -- HDim7 -> [1,  0,  0,  1,  0,  0,  1,  0,  0,  0,  1,   0 ]   
  -- Dim   -> [1,  0,  0,  1,  0,  0,  1,  0,  0,  0,  0,   0 ]   
  -- Maj6  -> [1,  0,  0,  0,  1,  0,  0,  1,  0,  1,  0,   0 ] 
  Sev   -> [1,  0,  0,  0,  1,  0,  0,  1,  0,  0,  1,   0 ] 
  Dim7  -> [1,  0,  0,  1,  0,  0,  1,  0,  0,  1,  0,   0 ] 
  -- Maj7  -> [1,  0,  0,  0,  1,  0,  0,  1,  0,  0,  0,   1 ] 
  -- Min7  -> [1,  0,  0,  1,  0,  0,  0,  1,  0,  0,  1,   0 ]
  -- Min6  -> [1,  0,  0,  1,  0,  0,  0,  1,  0,  1,  0,   0 ]
  -- Sus4  -> [1,  0,  0,  0,  0,  1,  0,  1,  0,  0,  0,   0 ] 
  
  None  -> [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0 ] 
  -- None -> [1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,   1 ] 
  _     -> []
  -- none
  -- all
  -- ....
         
--------------------------------------------------------------------------------
-- Chroma key estimation
--------------------------------------------------------------------------------
         
beatSyncKeyStrenth :: BeatTrackerData -> ChordinoData -> [[NumData]]
beatSyncKeyStrenth bts key = 
  map matchKeyProfiles . meanBeatSyncVectors {- . mergeBeats 8 -} $ beatSync bts key

nanToZero :: RealFloat a => a -> a
nanToZero n = if isNaN n then 0 else n

-- printKeyStrength :: BeatTrackerData -> ChordinoData -> IO ()
-- printKeyStrength bts chrm = 
  -- let showLn :: [NumData] -> IO ()
      -- showLn x = do mapM_ (putStr . printf "%.2f ") x ; putStr "\n"
  -- in mapM_ showLn $ beatSyncKeyStrenth bts chrm
      
matchKeyProfiles :: Vector NumData -> [NumData]
-- matchKeyProfiles chroma = map (\x -> pnorm PNorm2 (chroma - x)) allKeyProfiles
matchKeyProfiles crm = map (nanToZero . GSL.correlation crm) allKeyProfiles

allKeyProfiles :: [Vector NumData]
allKeyProfiles = map (fromList . keyToProfile) keyMap   
  
keyToProfile :: Key -> [NumData]          
keyToProfile (Key root m) = reverseShift (diaNatToSemi root) (selectProfile m) 

selectProfile :: Mode -> [NumData]
selectProfile MajMode = krumhanslProfCMaj
selectProfile MinMode = krumhanslProfCMin          
          
krumhanslProfCMaj, krumhanslProfCMin :: [NumData]
krumhanslProfCMaj = 
  [6.35, 2.23, 3.48, 2.33, 4.38, 4.09, 2.52, 5.19, 2.39, 3.66, 2.29, 2.88]
krumhanslProfCMin = 
  [6.33, 2.68, 3.52, 5.38, 2.60, 3.53, 2.54, 4.75, 3.98, 2.69, 3.34, 3.17]
  
--------------------------------------------------------------------------------
-- key strength Matrix Computations
--------------------------------------------------------------------------------
                             
toBeatSyncKey :: BeatTrackerData -> KeyStrengthData -> [Key]
toBeatSyncKey bts key = 
  medianFilter 4 . map ((!!) keyMap . maxIndex) . meanBeatSyncVectors 
  $ beatSync bts key

-- returns a (median based) key label given a Nx24 key strength matrix
findKeyIndex :: Matrix NumData -> Key
findKeyIndex = (!!) keyMap . median .  map maxIndex . toRows

toKeyMatrix :: KeyStrengthData -> Matrix NumData
toKeyMatrix = fromLists . map mergeKeyData where
  mergeKeyData :: ChordinoLine -> [NumData]
  mergeKeyData (ChordinoLine _ ma mi) = ma ++ mi

-- findKeyLab :: Matrix NumData -> Key
-- findKeyLab m = keyMap !! (findKeyIndex m)

-- the chroma map uses a circle of fifths based ordering
keyMap :: [Key]
keyMap =  [ Key (Note (Just Sh) F) MajMode -- "F#" 6
          , Key (Note Nothing   B) MajMode -- "B"  11
          , Key (Note Nothing   E) MajMode -- "E"  4
          , Key (Note Nothing   A) MajMode -- "A"  9
          , Key (Note Nothing   D) MajMode -- "D"  2
          , Key (Note Nothing   G) MajMode -- "G"  7
          , Key (Note Nothing   C) MajMode -- "C"  0
          , Key (Note Nothing   F) MajMode -- "F"  5
          , Key (Note (Just Fl) B) MajMode -- "Bb" 10
          , Key (Note (Just Fl) E) MajMode -- "Eb" 3
          , Key (Note (Just Fl) A) MajMode -- "Ab" 8
          , Key (Note (Just Fl) D) MajMode -- "Db" 1
          , Key (Note (Just Fl) E) MinMode -- "Ebm" 3
          , Key (Note (Just Sh) G) MinMode -- "G#m" 8
          , Key (Note (Just Sh) C) MinMode -- "C#m" 1
          , Key (Note (Just Sh) F) MinMode -- "F#m" 6
          , Key (Note Nothing   B) MinMode -- "Bm"  11
          , Key (Note Nothing   E) MinMode -- "Em"  4
          , Key (Note Nothing   A) MinMode -- "Am"  9
          , Key (Note Nothing   D) MinMode -- "Dm"  2
          , Key (Note Nothing   G) MinMode -- "Gm"  7
          , Key (Note Nothing   C) MinMode -- "Cm"  0
          , Key (Note Nothing   F) MinMode -- "Fm"  5
          , Key (Note (Just Fl) B) MinMode -- "Bbm" 10
          ]

--------------------------------------------------------------------------------
-- key annotation preparation
--------------------------------------------------------------------------------

-- syncronises the last beat and key frame to match the last chord frame
syncWithAnnKey ::  [ChordinoLine] -> [NumData] -> [TimedData Key] 
               -> ([NumData], [TimedData Key])
syncWithAnnKey chrm beat key = 
  let endTime   = time $ last chrm 
      beat'     = takeWhile (< endTime) beat ++ [endTime]
      none (Key r _) = r /= Note Nothing N -- filter the None keys
      noNoneKey = (filter (none . getData) key)
      -- reset key start timestamp to 0.0 and end timestampt to chorma end
      resetHead = (head noNoneKey) {onset = 0.0} : tail noNoneKey
      (l,[lst]) = splitAt ((length resetHead) - 1) resetHead
      -- key' = sampleAt beat' (l ++[lst{offset = endTime}])
  in  (beat',l ++[lst{offset = endTime}])
  
--------------------------------------------------------------------------------
-- general Matrix stuff
--------------------------------------------------------------------------------

-- given a matrix, calculates the mean vector 
mean :: (Product t, Fractional t) => Matrix t -> Vector t
mean a = constant (recip . fromIntegral . rows $ a) (rows a) <> a
